import { MediaSet } from './generated/media_set';
import { stateReducer, State, PlayState } from './AppState';
import { from } from 'rxjs';
import { CanvasWidth, SelectionMode } from './HudCanvasState';

const initialState: State = {
  selection: { start: 0, end: 0 },
  viewport: { start: 0, end: 441000 },
  overviewPeaks: from([]),
  waveformPeaks: from([]),
  selectionCanvas: { x1: 0, x2: 0 },
  viewportCanvas: { x1: 0, x2: CanvasWidth },
  position: { currentTime: 0, frame: 0, percent: 0 },
  audioSrc: '',
  videoSrc: '',
  currentTime: undefined,
  playState: PlayState.Paused,
};

describe('stateReducer', () => {
  describe('audiosourceloaded', () => {
    describe.each([
      {
        src: 'foo.opus',
        numFrames: 22050,
        wantViewport: { start: 0, end: 1100 },
        wantViewportCanvas: { x1: 0, x2: 100 },
      },
      {
        src: 'foo.opus',
        numFrames: 44100000,
        wantViewport: { start: 0, end: 2205000 },
        wantViewportCanvas: { x1: 0, x2: 100 },
      },
    ])(
      '$numFrames frames',
      ({ src, numFrames, wantViewport, wantViewportCanvas }) => {
        it('generates the expected state', () => {
          const mediaSet = MediaSet.fromPartial({
            id: '123',
            audioFrames: 0,
          });
          const state = stateReducer(
            { ...initialState, mediaSet },
            { type: 'audiosourceloaded', src, numFrames }
          );

          expect(state.mediaSet!.audioFrames).toEqual(numFrames);
          expect(state.audioSrc).toEqual(src);
          expect(state.viewport).toEqual(wantViewport);
          expect(state.viewportCanvas).toEqual(wantViewportCanvas);
        });
      }
    );
  });

  describe('setviewport', () => {
    describe.each([
      {
        audioFrames: 441000,
        viewport: { start: 0, end: 44100 },
        selection: { start: 0, end: 0 },
        wantViewportCanvas: { x1: 0, x2: 200 },
        wantSelectionCanvas: { x1: 0, x2: 0 },
      },
      {
        audioFrames: 441000,
        viewport: { start: 0, end: 441000 },
        selection: { start: 0, end: 0 },
        wantViewportCanvas: { x1: 0, x2: 2000 },
        wantSelectionCanvas: { x1: 0, x2: 0 },
      },
      {
        audioFrames: 441000,
        viewport: { start: 0, end: 441000 },
        selection: { start: 0, end: 44100 },
        wantViewportCanvas: { x1: 0, x2: 2000 },
        wantSelectionCanvas: { x1: 0, x2: 200 },
      },
      {
        audioFrames: 441000,
        viewport: { start: 0, end: 22050 },
        selection: { start: 0, end: 44100 },
        wantViewportCanvas: { x1: 0, x2: 100 },
        wantSelectionCanvas: { x1: 0, x2: 2000 },
      },
      {
        audioFrames: 441000,
        viewport: { start: 44100, end: 88200 },
        selection: { start: 22050, end: 66150 },
        wantViewportCanvas: { x1: 200, x2: 400 },
        wantSelectionCanvas: { x1: 0, x2: 1000 },
      },
    ])(
      'selection $selection.start-$selection.end, viewport: $viewport.start-$viewport.end',
      ({
        audioFrames,
        viewport,
        selection,
        wantViewportCanvas,
        wantSelectionCanvas,
      }) => {
        it('generates the expected state', () => {
          const mediaSet = MediaSet.fromPartial({
            id: '123',
            audioFrames: audioFrames,
          });
          const state = stateReducer(
            { ...initialState, mediaSet: mediaSet, selection: selection },
            { type: 'setviewport', viewport: viewport }
          );

          expect(state.viewport).toEqual(viewport);
          expect(state.selection).toEqual(selection);
          expect(state.viewportCanvas).toEqual(wantViewportCanvas);
          expect(state.selectionCanvas).toEqual(wantSelectionCanvas);
        });
      }
    );
  });

  describe('viewportchanged', () => {
    describe.each([
      {
        audioFrames: 441000,
        event: {
          mode: SelectionMode.Selecting,
          prevMode: SelectionMode.Selecting,
          selection: { x1: 0, x2: 200 },
        },
        selection: { start: 0, end: 0 },
        wantViewport: { start: 0, end: 441000 },
        wantSelectionCanvas: { x1: 0, x2: 0 },
      },
      {
        audioFrames: 441000,
        event: {
          mode: SelectionMode.Normal,
          prevMode: SelectionMode.Selecting,
          selection: { x1: 0, x2: 200 },
        },
        selection: { start: 0, end: 0 },
        wantViewport: { start: 0, end: 44100 },
        wantSelectionCanvas: { x1: 0, x2: 0 },
      },
      {
        audioFrames: 441000,
        event: {
          mode: SelectionMode.Normal,
          prevMode: SelectionMode.Selecting,
          selection: { x1: 0, x2: 200 },
        },
        selection: { start: 0, end: 22050 },
        wantViewport: { start: 0, end: 44100 },
        wantSelectionCanvas: { x1: 0, x2: 1000 },
      },
      {
        audioFrames: 441000,
        event: {
          mode: SelectionMode.Normal,
          prevMode: SelectionMode.Selecting,
          selection: { x1: 1000, x2: 1500 },
        },
        selection: { start: 220500, end: 264600 },
        wantViewport: { start: 220500, end: 330750 },
        wantSelectionCanvas: { x1: 0, x2: 800 },
      },
    ])(
      'mode $event.mode, audioFrames $audioFrames, canvas range $event.selection.x1-$event.selection.x2, selectedFrames $selection.start-$selection.end',
      ({
        audioFrames,
        event,
        selection,
        wantViewport,
        wantSelectionCanvas,
      }) => {
        it('generates the expected state', () => {
          const mediaSet = MediaSet.fromPartial({
            id: '123',
            audioFrames: audioFrames,
          });
          const state = stateReducer(
            {
              ...initialState,
              mediaSet: mediaSet,
              viewport: { start: 0, end: audioFrames },
              selection: selection,
            },
            { type: 'viewportchanged', event }
          );

          expect(state.selection).toEqual(selection);
          expect(state.viewport).toEqual(wantViewport);
          expect(state.selectionCanvas).toEqual(wantSelectionCanvas);
        });
      }
    );
  });

  describe('waveformselectionchanged', () => {
    describe.each([
      {
        name: 'paused',
        audioSampleRate: 44100,
        event: {
          mode: SelectionMode.Selecting,
          prevMode: SelectionMode.Selecting,
          selection: { x1: 100, x2: 200 },
        },
        playState: PlayState.Paused,
        position: { frame: 0, currentTime: 0, percent: 0 },
        viewport: { start: 0, end: 88200 },
        wantSelection: { start: 4410, end: 8820 },
        wantCurrentTime: undefined,
      },
      {
        name: 'playing, viewport 100%, selection is in progress',
        audioSampleRate: 44100,
        event: {
          mode: SelectionMode.Selecting,
          prevMode: SelectionMode.Selecting,
          selection: { x1: 200, x2: 220 },
        },
        playState: PlayState.Playing,
        position: { frame: 22000, currentTime: 0.4988, percent: 4.98 },
        viewport: { start: 0, end: 441000 },
        wantSelection: { start: 44100, end: 48510 },
        wantCurrentTime: undefined,
      },
      {
        name: 'playing, viewport partial, selection is in progress',
        audioSampleRate: 44100,
        event: {
          mode: SelectionMode.Selecting,
          prevMode: SelectionMode.Selecting,
          selection: { x1: 0, x2: 100 },
        },
        playState: PlayState.Playing,
        position: { frame: 22000, currentTime: 0.4988, percent: 4.98 },
        viewport: { start: 88200, end: 176400 },
        wantSelection: { start: 88200, end: 92610 },
        wantCurrentTime: undefined,
      },
      {
        name: 'playing, selection is ending, currFrame is before selection start',
        audioSampleRate: 44100,
        event: {
          mode: SelectionMode.Normal,
          prevMode: SelectionMode.Selecting,
          selection: { x1: 1100, x2: 1200 },
        },
        playState: PlayState.Playing,
        position: { frame: 22000, currentTime: 0.4988, percent: 4.98 },
        viewport: { start: 0, end: 88200 },
        wantSelection: { start: 48510, end: 52920 },
        wantCurrentTime: 1.1,
      },
      {
        name: 'playing, selection is ending, currFrame is within selection',
        audioSampleRate: 44100,
        event: {
          mode: SelectionMode.Normal,
          prevMode: SelectionMode.Selecting,
          selection: { x1: 1001, x2: 1200 },
        },
        playState: PlayState.Playing,
        position: { frame: 50000, currentTime: 1.133, percent: 11.33 },
        viewport: { start: 0, end: 88200 },
        wantSelection: { start: 44144, end: 52920 },
        wantCurrentTime: undefined,
      },
      {
        name: 'playing, selection is ending, currFrame is after selection end',
        audioSampleRate: 44100,
        event: {
          mode: SelectionMode.Normal,
          prevMode: SelectionMode.Selecting,
          selection: { x1: 1001, x2: 1200 },
        },
        playState: PlayState.Playing,
        position: { frame: 88200, currentTime: 2.0, percent: 20.0 },
        viewport: { start: 0, end: 88200 },
        wantSelection: { start: 44144, end: 52920 },
        wantCurrentTime: 1.000997732426304,
      },
    ])(
      '$name',
      ({
        audioSampleRate,
        event,
        playState,
        position,
        viewport,
        wantSelection,
        wantCurrentTime,
      }) => {
        it('generates the expected state', () => {
          const mediaSet = MediaSet.fromPartial({
            id: '123',
            audioFrames: 441000,
            audioSampleRate: audioSampleRate,
          });
          const state = stateReducer(
            {
              ...initialState,
              position,
              mediaSet,
              playState,
              viewport,
            },
            { type: 'waveformselectionchanged', event }
          );

          expect(state.selection).toEqual(wantSelection);
          expect(state.currentTime).toEqual(wantCurrentTime);
        });
      }
    );
  });

  describe('positionchanged', () => {
    describe.each([
      {
        name: 'playing, 48k',
        audioSampleRate: 48000,
        newCurrentTime: 0.02,
        position: { frame: 0, currentTime: 0, percent: 0 },
        selection: { start: 0, end: 0 },
        wantPosition: {
          frame: 960,
          currentTime: 0.02,
          percent: 0.21768707482993196,
        },
        wantCurrentTime: undefined,
        wantPlayState: PlayState.Playing,
      },
      {
        name: 'playing, 44.1k',
        audioSampleRate: 44100,
        newCurrentTime: 8.51,
        position: { frame: 360000, currentTime: 8.16, percent: 81.6 },
        selection: { start: 0, end: 0 },
        wantPosition: { frame: 375291, currentTime: 8.51, percent: 85.1 },
        wantCurrentTime: undefined,
        wantPlayState: PlayState.Playing,
      },
      {
        name: 'playing, passed selection end',
        audioSampleRate: 44100,
        newCurrentTime: 8.51,
        position: { frame: 360000, currentTime: 8.16, percent: 81.6 },
        selection: { start: 22050, end: 375290 },
        wantPosition: { frame: 375291, currentTime: 8.51, percent: 85.1 },
        wantCurrentTime: 0.5,
        wantPlayState: PlayState.Paused,
      },
    ])(
      '$name',
      ({
        audioSampleRate,
        newCurrentTime,
        position,
        selection,
        wantPosition,
        wantCurrentTime,
        wantPlayState,
      }) => {
        it('generates the expected state', () => {
          const mediaSet = MediaSet.fromPartial({
            id: '123',
            audioFrames: 441000,
            audioSampleRate: audioSampleRate,
          });
          const state = stateReducer(
            {
              ...initialState,
              playState: PlayState.Playing,
              mediaSet,
              position,
              selection,
            },
            { type: 'positionchanged', currentTime: newCurrentTime }
          );

          expect(state.position).toEqual(wantPosition);
          expect(state.playState).toEqual(wantPlayState);
          expect(state.currentTime).toEqual(wantCurrentTime);
        });
      }
    );
  });
});
